@@ -29,8 +29,10 @@
 #include "UrlDecode.h"
 
 #include "BasicUI.h"
+#include "ExportUtils.h"
 
 #include "StringUtils.h"
+#include "UrlEncode.h"
 
 namespace audacity::cloud::audiocom
 {
@@ -43,17 +45,18 @@ const std::string_view uriPrefix = "audacity://link";
 const std::string_view usernamePrefix = "username=";
 const std::string_view passwordPrefix = "password=";
 const std::string_view tokenPrefix = "token=";
+const std::string_view authClientPrefix = "authclient=";
+const std::string_view responseTypePrefix = "response_type=";
+const std::string_view clientIdPrefix = "client_id=";
 const std::string_view authorizationCodePrefix = "authorization_code=";
+const std::string_view codePrefix = "code=";
+const std::string_view urlPrefix = "url=";
+const std::string_view userPrefix = "user=";
 
-void WriteCommonFields(
-   rapidjson::Document& document, std::string_view grantType, std::string_view scope)
+void WriteClientFields(rapidjson::Document& document)
 {
    using namespace rapidjson;
 
-   document.AddMember(
-      "grant_type", StringRef(grantType.data(), grantType.size()),
-      document.GetAllocator());
-
    const auto clientID = GetServiceConfig().GetOAuthClientID();
    const auto clientSecret = GetServiceConfig().GetOAuthClientSecret();
 
@@ -66,15 +69,48 @@ void WriteCommonFields(
       "client_secret",
       Value(clientSecret.data(), clientSecret.size(), document.GetAllocator()),
       document.GetAllocator());
+}
+
+void WriteAccessFields(rapidjson::Document& document, std::string_view grantType, std::string_view scope)
+{
+   using namespace rapidjson;
+
+   document.AddMember(
+      "grant_type", StringRef(grantType.data(), grantType.size()),
+      document.GetAllocator());
 
    document.AddMember(
       "scope", StringRef(scope.data(), scope.size()), document.GetAllocator());
 }
 
+void WriteCommonFields(
+   rapidjson::Document& document, std::string_view grantType, std::string_view scope)
+{
+   WriteClientFields(document);
+   WriteAccessFields(document, grantType, scope);
+}
+
+template<typename Elem, typename First, typename ...Others>
+void append(std::basic_string<Elem>& dest, First&& first, Others&& ...others)
+{
+   dest.append(first);
+   if constexpr (sizeof...(others) != 0)
+      append(dest, std::forward<Others>(others)...);
+}
+
+template<typename First, typename ...Others>
+auto concat(First&& first, Others&& ...others)
+{
+   std::basic_string<typename First::value_type> dest(first);
+   append(dest, std::forward<Others>(others)...);
+   return dest;
+}
+
 } // namespace
 
 void OAuthService::ValidateAuth(
-   std::function<void(std::string_view)> completedHandler, bool silent)
+   std::function<void(std::string_view)> completedHandler, AudiocomTrace trace,
+   bool silent)
 {
    if (HasAccessToken() || !HasRefreshToken())
    {
@@ -83,11 +119,13 @@ void OAuthService::ValidateAuth(
       return;
    }
 
-   AuthoriseRefreshToken(GetServiceConfig(), std::move(completedHandler), silent);
+   AuthoriseRefreshToken(
+      GetServiceConfig(), trace, std::move(completedHandler), silent);
 }
 
 bool OAuthService::HandleLinkURI(
-   std::string_view uri, std::function<void(std::string_view)> completedHandler)
+   std::string_view uri, AudiocomTrace trace,
+   std::function<void(std::string_view)> completedHandler)
 {
    if (!IsPrefixedInsensitive(uri, uriPrefix))
    {
@@ -97,7 +135,7 @@ bool OAuthService::HandleLinkURI(
    }
 
    // It was observed, that sometimes link is passed as audacity://link/
-   // This is valid from URI point of view, but we need to handle it separately
+   // This is valid trace URI point of view, but we need to handle it separately
    const auto argsStart = uri.find("?");
 
    if (argsStart == std::string_view::npos)
@@ -114,6 +152,7 @@ bool OAuthService::HandleLinkURI(
    std::string_view username;
    std::string_view password;
    std::string_view authorizationCode;
+   auto useAudioComRedirectURI = false;
 
    while (!args.empty())
    {
@@ -129,26 +168,39 @@ bool OAuthService::HandleLinkURI(
       else if (IsPrefixed(arg, tokenPrefix))
          token = arg.substr(tokenPrefix.length());
       else if (IsPrefixed(arg, authorizationCodePrefix))
+      {
+         //authorization code was generated for audio.com, not audacity...
+         useAudioComRedirectURI = true;
          authorizationCode = arg.substr(authorizationCodePrefix.length());
+      }
+      else if (IsPrefixed(arg, codePrefix))
+         authorizationCode = arg.substr(codePrefix.length());
+   }
+
+   // Some browsers (safari) add an extra trailing chars we don't need
+   size_t hashPos = authorizationCode.find('#');
+    
+   if (hashPos != std::string::npos) {
+      authorizationCode = authorizationCode.substr(0, hashPos);
    }
 
    // We have a prioritized list of authorization methods
    if (!authorizationCode.empty())
    {
       AuthoriseCode(
-         GetServiceConfig(), authorizationCode, std::move(completedHandler));
+         GetServiceConfig(), authorizationCode, useAudioComRedirectURI, trace,
+         std::move(completedHandler));
    }
    else if (!token.empty())
    {
       AuthoriseRefreshToken(
-         GetServiceConfig(), token, std::move(completedHandler), false);
+         GetServiceConfig(), token, trace, std::move(completedHandler), false);
    }
    else if (!username.empty() && !password.empty())
    {
       AuthorisePassword(
-         GetServiceConfig(),
-         audacity::UrlDecode(std::string(username)),
-         audacity::UrlDecode(std::string(password)),
+         GetServiceConfig(), audacity::UrlDecode(std::string(username)),
+         audacity::UrlDecode(std::string(password)), trace,
          std::move(completedHandler));
    }
    else
@@ -162,7 +214,7 @@ bool OAuthService::HandleLinkURI(
    return true;
 }
 
-void OAuthService::UnlinkAccount()
+void OAuthService::UnlinkAccount(AudiocomTrace trace)
 {
    std::lock_guard<std::recursive_mutex> lock(mMutex);
 
@@ -172,12 +224,12 @@ void OAuthService::UnlinkAccount()
 
    // Unlink account is expected to be called only
    // on UI thread
-   Publish({ {}, {}, false });
+   Publish({ {}, {}, trace, false });
 }
 
 void OAuthService::AuthorisePassword(
    const ServiceConfig& config, std::string_view userName,
-   std::string_view password,
+   std::string_view password, AudiocomTrace trace,
    std::function<void(std::string_view)> completedHandler)
 {
    using namespace rapidjson;
@@ -200,12 +252,12 @@ void OAuthService::AuthorisePassword(
    document.Accept(writer);
 
    DoAuthorise(
-      config, { buffer.GetString(), buffer.GetSize() },
+      config, { buffer.GetString(), buffer.GetSize() }, trace,
       std::move(completedHandler), false);
 }
 
 void OAuthService::AuthoriseRefreshToken(
-   const ServiceConfig& config, std::string_view token,
+   const ServiceConfig& config, std::string_view token, AudiocomTrace trace,
    std::function<void(std::string_view)> completedHandler, bool silent)
 {
    using namespace rapidjson;
@@ -224,24 +276,24 @@ void OAuthService::AuthoriseRefreshToken(
    document.Accept(writer);
 
    DoAuthorise(
-      config, { buffer.GetString(), buffer.GetSize() },
+      config, { buffer.GetString(), buffer.GetSize() }, trace,
       std::move(completedHandler), silent);
 }
 
 void OAuthService::AuthoriseRefreshToken(
-   const ServiceConfig& config,
+   const ServiceConfig& config, AudiocomTrace trace,
    std::function<void(std::string_view)> completedHandler, bool silent)
 {
    std::lock_guard<std::recursive_mutex> lock(mMutex);
 
    AuthoriseRefreshToken(
-      config, audacity::ToUTF8(refreshToken.Read()),
+      config, audacity::ToUTF8(refreshToken.Read()), trace,
       std::move(completedHandler), silent);
 }
 
 void OAuthService::AuthoriseCode(
-   const ServiceConfig& config, std::string_view authorizationCode,
-   std::function<void(std::string_view)> completedHandler)
+   const ServiceConfig& config, std::string_view authorizationCode, bool useAudioComRedirectURI,
+   AudiocomTrace trace, std::function<void(std::string_view)> completedHandler)
 {
    using namespace rapidjson;
 
@@ -254,7 +306,7 @@ void OAuthService::AuthoriseCode(
       "code", StringRef(authorizationCode.data(), authorizationCode.size()),
       document.GetAllocator());
 
-   const auto redirectURI = config.GetOAuthRedirectURL();
+   const auto redirectURI = useAudioComRedirectURI ? config.GetOAuthRedirectURL() : std::string("audacity://link");
 
    document.AddMember(
       "redirect_uri", StringRef(redirectURI.data(), redirectURI.size()),
@@ -265,7 +317,7 @@ void OAuthService::AuthoriseCode(
    document.Accept(writer);
 
    DoAuthorise(
-      config, { buffer.GetString(), buffer.GetSize() },
+      config, { buffer.GetString(), buffer.GetSize() }, trace,
       std::move(completedHandler), false);
 }
 
@@ -290,9 +342,140 @@ std::string OAuthService::GetAccessToken() const
    return {};
 }
 
+std::string OAuthService::MakeOAuthRequestURL(std::string_view authClientId)
+{
+   using namespace audacity::network_manager;
+
+   return concat(
+      GetServiceConfig().GetAPIUrl("/auth/authorize?"),
+      authClientPrefix, authClientId, "&",
+      responseTypePrefix, "code", "&",
+      clientIdPrefix, GetServiceConfig().GetOAuthClientID(),
+      "&redirect_uri=audacity://link"
+   );
+}
+
+std::string OAuthService::MakeAudioComAuthorizeURL(std::string_view userId, std::string_view redirectUrl)
+{
+   auto token = GetAccessToken();
+
+   // Remove token type from the token string
+   size_t pos = token.find(' ');
+   if (pos != std::string::npos) {
+      token = token.substr(pos + 1);
+   }
+
+   return concat(
+      GetServiceConfig().GetAuthWithRedirectURL(), "?",
+      tokenPrefix, token, "&",
+      userPrefix, userId, "&",
+      urlPrefix, redirectUrl
+   );
+}
+
+void OAuthService::Authorize(std::string_view email,
+                             std::string_view password,
+                             AuthSuccessCallback onSuccess,
+                             AuthFailureCallback onFailure,
+                             AudiocomTrace trace)
+{
+   using namespace audacity::network_manager;
+   using rapidjson::StringRef;
+
+   rapidjson::Document document;
+   document.SetObject();
+
+   WriteCommonFields(document, "password", "all");
+
+   document.AddMember(
+      "username", StringRef(email.data(), email.size()),
+      document.GetAllocator());
+
+   document.AddMember(
+      "password", StringRef(password.data(), password.size()),
+      document.GetAllocator());
+
+   rapidjson::StringBuffer buffer;
+   rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
+   document.Accept(writer);
+
+   Request request(GetServiceConfig().GetAPIUrl("/auth/token"));
+   request.setHeader(common_headers::ContentType, common_content_types::ApplicationJson);
+   request.setHeader(common_headers::Accept, common_content_types::ApplicationJson);
+
+   auto response = NetworkManager::GetInstance().doPost(request,  buffer.GetString(), buffer.GetSize());
+   response->setRequestFinishedCallback(
+      [response, this, trace,
+         onSuccess = std::move(onSuccess),
+         onFailure = std::move(onFailure)](auto) mutable
+      {
+         const auto httpCode = response->getHTTPCode();
+         const auto body = response->readAll<std::string>();
+         if(httpCode == 200)
+            ParseTokenResponse(body, std::move(onSuccess), std::move(onFailure), trace, false);
+         else
+         {
+            if(onFailure)
+               onFailure(httpCode, body);
+
+            SafePublish({ {}, body, trace, false, false });
+         }
+      });
+}
+
+void OAuthService::Register(std::string_view email,
+                            std::string_view password,
+                            AuthSuccessCallback successCallback,
+                            AuthFailureCallback failureCallback,
+                            AudiocomTrace trace)
+{
+   using namespace audacity::network_manager;
+   using rapidjson::StringRef;
+
+   rapidjson::Document document;
+   document.SetObject();
+
+   WriteClientFields(document);
+
+   document.AddMember(
+      "email", StringRef(email.data(), email.size()),
+      document.GetAllocator());
+
+   document.AddMember(
+      "password", StringRef(password.data(), password.size()),
+      document.GetAllocator());
+
+   rapidjson::StringBuffer buffer;
+   rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
+   document.Accept(writer);
+
+   Request request(GetServiceConfig().GetAPIUrl("/auth/register"));
+   request.setHeader(common_headers::ContentType, common_content_types::ApplicationJson);
+   request.setHeader(common_headers::Accept, common_content_types::ApplicationJson);
+
+   auto response = NetworkManager::GetInstance().doPost(request,  buffer.GetString(), buffer.GetSize());
+   response->setRequestFinishedCallback(
+      [response, this, trace,
+          successCallback = std::move(successCallback),
+          failureCallback = std::move(failureCallback)] (auto) mutable
+      {
+         const auto httpCode = response->getHTTPCode();
+         const auto body = response->readAll<std::string>();
+         if(httpCode == 200)
+            ParseTokenResponse(body, std::move(successCallback), std::move(failureCallback), trace, false);
+         else
+         {
+            if(failureCallback)
+               failureCallback(httpCode, body);
+
+            SafePublish({ {}, body, trace, false, false });
+         }
+      });
+}
+
 void OAuthService::DoAuthorise(
-   const ServiceConfig& config, std::string_view payload,
-   std::function<void(std::string_view)> completedHandler, bool silent)
+   const ServiceConfig& config, std::string_view payload, AudiocomTrace trace,
+   AuthSuccessCallback completedHandler, bool silent)
 {
    using namespace audacity::network_manager;
 
@@ -308,8 +491,8 @@ void OAuthService::DoAuthorise(
       request, payload.data(), payload.size());
 
    response->setRequestFinishedCallback(
-      [response, this, handler = std::move(completedHandler), silent](auto)
-      {
+      [response, this, handler = std::move(completedHandler), silent,
+       trace](auto) mutable {
          const auto httpCode = response->getHTTPCode();
          const auto body = response->readAll<std::string>();
 
@@ -320,56 +503,64 @@ void OAuthService::DoAuthorise(
 
             // Token has expired?
             if (httpCode == 422)
-               BasicUI::CallAfter([this] { UnlinkAccount(); });
+               BasicUI::CallAfter([this, trace] { UnlinkAccount(trace); });
             else
-               SafePublish({ {}, body, false, silent });
+               SafePublish({ {}, body, trace, false, silent });
 
             return;
          }
-
-         rapidjson::Document document;
-         document.Parse(body.data(), body.size());
-
-         if (!document.IsObject())
-         {
-            if (handler)
-               handler({});
-
-            SafePublish({ {}, body, false, silent });
-            return;
-         }
-
-         const auto tokenType = document["token_type"].GetString();
-         const auto accessToken = document["access_token"].GetString();
-         const auto expiresIn = document["expires_in"].GetInt64();
-         const auto newRefreshToken = document["refresh_token"].GetString();
-
-         {
-            std::lock_guard<std::recursive_mutex> lock(mMutex);
-
-            mAccessToken = std::string(tokenType) + " " + accessToken;
-            mTokenExpirationTime =
-               Clock::now() + std::chrono::seconds(expiresIn);
-         }
-
-         BasicUI::CallAfter(
-            [token = std::string(newRefreshToken)]()
-            {
-               // At this point access token is already written,
-               // only refresh token is updated.
-               refreshToken.Write(token);
-               gPrefs->Flush();
-            });
-
-         if (handler)
-            handler(mAccessToken);
-
-         // The callback only needs the access token, so invoke it immediately.
-         // Networking is thread safe
-         SafePublish({ mAccessToken, {}, true, silent });
+         ParseTokenResponse(body, std::move(handler), {}, trace, silent);
       });
 }
 
+void OAuthService::ParseTokenResponse(std::string_view body,
+                                      AuthSuccessCallback successCallback,
+                                      AuthFailureCallback failureCallback,
+                                      AudiocomTrace trace,
+                                      bool silent)
+{
+   rapidjson::Document document;
+   document.Parse(body.data(), body.size());
+
+   if (!document.IsObject())
+   {
+      if (failureCallback)
+         failureCallback(200, body);
+
+      SafePublish({ {}, body, trace, false, silent });
+      return;
+   }
+
+   const auto tokenType = document["token_type"].GetString();
+   const auto accessToken = document["access_token"].GetString();
+   const auto expiresIn = document["expires_in"].GetInt64();
+   const auto newRefreshToken = document["refresh_token"].GetString();
+
+   {
+      std::lock_guard<std::recursive_mutex> lock(mMutex);
+
+      mAccessToken = std::string(tokenType) + " " + accessToken;
+      mTokenExpirationTime =
+         Clock::now() + std::chrono::seconds(expiresIn);
+   }
+
+   BasicUI::CallAfter(
+      [token = std::string(newRefreshToken)]()
+      {
+         // At this point access token is already written,
+         // only refresh token is updated.
+         refreshToken.Write(token);
+         gPrefs->Flush();
+      });
+
+   if (successCallback)
+      successCallback(mAccessToken);
+
+   // The callback only needs the access token, so invoke it immediately.
+   // Networking is thread safe
+   SafePublish({ mAccessToken, {}, trace, true, silent });
+}
+
 void OAuthService::SafePublish(const AuthStateChangedMessage& message)
 {
    BasicUI::CallAfter([this, message]() { Publish(message); });
@@ -393,7 +584,7 @@ public:
 
    void OnSettingResetEnd() override
    {
-      GetOAuthService().UnlinkAccount();
+      GetOAuthService().UnlinkAccount(AudiocomTrace::ignore);
       refreshToken.Invalidate();
    }
 };
