@@ -11,6 +11,7 @@
 
 #include "UploadService.h"
 
+#include <cassert>
 #include <mutex>
 
 #include <wx/filefn.h>
@@ -23,18 +24,20 @@
 
 #include "OAuthService.h"
 #include "ServiceConfig.h"
+#include "UserService.h"
 
-#include "NetworkManager.h"
-#include "Request.h"
 #include "IResponse.h"
 #include "MultipartData.h"
+#include "NetworkManager.h"
+#include "NetworkUtils.h"
+#include "Request.h"
 
 #include "CodeConversions.h"
 
 #include "TempDirectory.h"
 #include "FileNames.h"
 
-namespace cloud::audiocom
+namespace audacity::cloud::audiocom
 {
 namespace
 {
@@ -50,7 +53,8 @@ std::string_view DeduceMimeType(const wxString& ext)
       return "audio/x-wav";
 }
 
-std::string GetUploadRequestPayload(const wxString& filePath, const wxString& projectName)
+std::string GetUploadRequestPayload(
+   const wxString& filePath, const wxString& projectName, bool isPublic)
 {
    rapidjson::Document document;
    document.SetObject();
@@ -88,6 +92,9 @@ std::string GetUploadRequestPayload(const wxString& filePath, const wxString& pr
       rapidjson::Value(static_cast<int64_t>(fileName.GetSize().GetValue())),
       document.GetAllocator());
 
+   document.AddMember(
+      "public", rapidjson::Value(isPublic), document.GetAllocator());
+
    rapidjson::StringBuffer buffer;
    rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
    document.Accept(writer);
@@ -111,6 +118,56 @@ std::string GetProgressPayload(uint64_t current, uint64_t total)
    return std::string(buffer.GetString());
 }
 
+UploadFailedPayload ParseUploadFailedMessage(const std::string& payloadText)
+{
+   rapidjson::StringStream stream(payloadText.c_str());
+   rapidjson::Document document;
+
+   document.ParseStream(stream);
+
+   if (!document.IsObject())
+   {
+      // This is unexpected, just return an empty object
+      assert(document.IsObject());
+      return {};
+   }
+
+   UploadFailedPayload payload;
+
+   auto readInt = [&document](const char* name) {
+      return document.HasMember(name) && document[name].IsInt() ?
+                document[name].GetInt() :
+                0;
+   };
+
+   auto readString = [&document](const char* name) -> const char*
+   {
+      return document.HasMember(name) && document[name].IsString() ?
+                document[name].GetString() :
+                "";
+   };
+
+   payload.code = readInt("code");
+   payload.status = readInt("status");
+
+   payload.name = readString("name");
+   payload.message = readString("message");
+
+   if (document.HasMember("errors") && document["errors"].IsObject())
+   {
+      for (auto& err : document["errors"].GetObject ())
+      {
+         if (!err.value.IsString())
+            continue;
+
+         payload.additionalErrors.emplace_back(
+            err.name.GetString(), err.value.GetString());
+      }
+   }
+
+   return payload;
+}
+
 
 // This class will capture itself inside the request handlers
 // by a strong reference. This way we ensure that it outlives all
@@ -121,22 +178,27 @@ struct AudiocomUploadOperation final :
 {
    AudiocomUploadOperation(
       const ServiceConfig& serviceConfig, wxString fileName,
-      wxString projectName,
+      wxString projectName, bool isPublic,
       UploadService::CompletedCallback completedCallback,
-      UploadService::ProgressCallback progressCallback)
+      UploadService::ProgressCallback progressCallback, AudiocomTrace trace)
        : mServiceConfig(serviceConfig)
        , mFileName(std::move(fileName))
        , mProjectName(std::move(projectName))
+       , mIsPublic(isPublic)
+       , mAudiocomTrace(trace)
        , mCompletedCallback(std::move(completedCallback))
        , mProgressCallback(std::move(progressCallback))
    {
    }
 
    const ServiceConfig& mServiceConfig;
-   
+
    const wxString mFileName;
    const wxString mProjectName;
 
+   const bool mIsPublic;
+
+   const AudiocomTrace mAudiocomTrace;
    UploadService::CompletedCallback mCompletedCallback;
    UploadService::ProgressCallback mProgressCallback;
 
@@ -148,6 +210,7 @@ struct AudiocomUploadOperation final :
 
    std::string mAudioID;
    std::string mUploadToken;
+   std::string mUserName;
 
    std::string mAudioSlug;
 
@@ -157,16 +220,23 @@ struct AudiocomUploadOperation final :
 
    mutable std::mutex mStatusMutex;
    mutable std::mutex mCallbacksMutex;
-   
+
    std::weak_ptr<audacity::network_manager::IResponse> mActiveResponse;
    bool mCompleted {};
    bool mAborted {};
 
-   void SetAuthHeader(audacity::network_manager::Request& request) const
+   void SetRequiredHeaders(audacity::network_manager::Request& request) const
    {
       if (!mAuthToken.empty())
          request.setHeader(
             audacity::network_manager::common_headers::Authorization, std::string(mAuthToken));
+
+      const auto language = mServiceConfig.GetAcceptLanguageValue();
+
+      if (!language.empty())
+         request.setHeader(
+            audacity::network_manager::common_headers::AcceptLanguage,
+            language);
    }
 
    void FailPromise(UploadOperationCompleted::Result result, std::string errorMessage)
@@ -181,7 +251,7 @@ struct AudiocomUploadOperation final :
       if (mCompletedCallback)
       {
          mCompletedCallback(
-            UploadOperationCompleted { result, std::move(errorMessage) });
+            UploadOperationCompleted { result, ParseUploadFailedMessage(errorMessage) });
       }
 
       mProgressCallback = {};
@@ -200,33 +270,43 @@ struct AudiocomUploadOperation final :
       if (mCompletedCallback)
       {
 
+         auto& userService = GetUserService();
+         auto& oauthService = GetOAuthService();
+
+         auto userId = audacity::ToUTF8(userService.GetUserId());
+         auto audioPage = mServiceConfig.GetAudioPagePath(
+                                      mUserName, mAudioSlug, mAudiocomTrace);
+
+         const auto uploadURL = mAuthToken.empty() ?
+                                   mServiceConfig.GetFinishUploadPage(
+                                      mAudioID, mUploadToken, mAudiocomTrace) :
+                                   oauthService.MakeAudioComAuthorizeURL(userId, audioPage);
+
          mCompletedCallback(
             { UploadOperationCompleted::Result::Success,
-              {},
-              mServiceConfig.GetFinishUploadPage(mAudioID, mUploadToken),
-              mAudioSlug });
+              UploadSuccessfulPayload { mAudioID, mAudioSlug, mUploadToken, uploadURL } });
       }
-      
+
       mProgressCallback = {};
       mCompletedCallback = {};
    }
 
    void InitiateUpload(std::string_view authToken)
-   {      
+   {
       using namespace audacity::network_manager;
 
       Request request(mServiceConfig.GetAPIUrl("/audio"));
-         
+
       request.setHeader(
          common_headers::ContentType, common_content_types::ApplicationJson);
-      
+
       request.setHeader(
          common_headers::Accept, common_content_types::ApplicationJson);
 
       mAuthToken = std::string(authToken);
-      SetAuthHeader(request);
+      SetRequiredHeaders(request);
 
-      const auto payload = GetUploadRequestPayload(mFileName, mProjectName);
+      const auto payload = GetUploadRequestPayload(mFileName, mProjectName, mIsPublic);
 
       std::lock_guard<std::mutex> lock(mStatusMutex);
 
@@ -242,7 +322,7 @@ struct AudiocomUploadOperation final :
       response->setRequestFinishedCallback(
          [response, sharedThis = shared_from_this(), this](auto) {
             auto responseCode = response->getHTTPCode();
-            
+
             if (responseCode == 201)
             {
                HandleUploadPolicy(response->readAll<std::string>());
@@ -271,7 +351,7 @@ struct AudiocomUploadOperation final :
    void HandleUploadPolicy(std::string uploadPolicyJSON)
    {
       using namespace audacity::network_manager;
-      
+
       rapidjson::Document document;
       document.Parse(uploadPolicyJSON.data(), uploadPolicyJSON.length());
 
@@ -315,10 +395,10 @@ struct AudiocomUploadOperation final :
          FailPromise(UploadOperationCompleted::Result::FileNotFound, {});
          return;
       }
-      
+
 
       const auto url = document["url"].GetString();
-      
+
       mSuccessUrl = document["success"].GetString();
       mFailureUrl = document["fail"].GetString();
       mProgressUrl = document["progress"].GetString();
@@ -332,6 +412,8 @@ struct AudiocomUploadOperation final :
 
          if (extra.HasMember("token"))
             mUploadToken = extra["token"].GetString();
+
+         mUserName = extra["audio"]["username"].GetString();
       }
 
       const auto encType = document.HasMember("enctype") ?
@@ -339,22 +421,22 @@ struct AudiocomUploadOperation final :
                               "multipart/form-data";
 
       Request request(url);
-      
+
       request.setHeader(common_headers::ContentType, encType);
       request.setHeader(
          common_headers::Accept, common_content_types::ApplicationJson);
 
       // We only lock late and for very short time
       std::lock_guard<std::mutex> lock(mStatusMutex);
-      
+
       if (mAborted)
          return;
-      
+
       auto response =
          NetworkManager::GetInstance().doPost(request, std::move(form));
 
       mActiveResponse = response;
-      
+
       response->setRequestFinishedCallback(
          [response, sharedThis = shared_from_this(), this](auto)
          {
@@ -371,11 +453,11 @@ struct AudiocomUploadOperation final :
    {
       {
          std::lock_guard<std::mutex> callbacksLock(mCallbacksMutex);
-         
+
          if (mProgressCallback)
             mProgressCallback(current, total);
       }
-      
+
       const auto now = Clock::now();
 
       if ((now - mLastProgressReportTime) > mServiceConfig.GetProgressCallbackTimeout())
@@ -384,7 +466,7 @@ struct AudiocomUploadOperation final :
 
          using namespace audacity::network_manager;
          Request request(mProgressUrl);
-         
+
          request.setHeader(
             common_headers::ContentType, common_content_types::ApplicationJson);
          request.setHeader(
@@ -396,7 +478,7 @@ struct AudiocomUploadOperation final :
 
          if (mAborted)
             return;
-         
+
          auto response = NetworkManager::GetInstance().doPatch(
             request, payload.data(), payload.size());
 
@@ -407,14 +489,14 @@ struct AudiocomUploadOperation final :
    void HandleS3UploadCompleted(std::shared_ptr<audacity::network_manager::IResponse> response)
    {
       using namespace audacity::network_manager;
-      
+
       const auto responseCode = response->getHTTPCode();
 
       const bool success =
          responseCode == 200 || responseCode == 201 || responseCode == 204;
-      
+
       Request request(success ? mSuccessUrl : mFailureUrl);
-      SetAuthHeader(request);
+      SetRequiredHeaders(request);
 
       std::lock_guard<std::mutex> lock(mStatusMutex);
 
@@ -461,7 +543,7 @@ struct AudiocomUploadOperation final :
 
          if (auto activeResponse = mActiveResponse.lock())
             activeResponse->abort();
-      } 
+      }
 
       std::lock_guard<std::mutex> callbacksLock(mCallbacksMutex);
 
@@ -476,7 +558,7 @@ struct AudiocomUploadOperation final :
    void DiscardResult() override
    {
       using namespace audacity::network_manager;
-      
+
       Abort();
 
       auto url = mServiceConfig.GetAPIUrl("/audio");
@@ -500,8 +582,9 @@ UploadService::UploadService(const ServiceConfig& config, OAuthService& service)
 }
 
 UploadOperationHandle UploadService::Upload(
-   const wxString& fileName, const wxString& projectName,
-   CompletedCallback completedCallback, ProgressCallback progressCallback)
+   const wxString& fileName, const wxString& projectName, bool isPublic,
+   CompletedCallback completedCallback, ProgressCallback progressCallback,
+   AudiocomTrace trace)
 {
    if (!wxFileExists(fileName))
    {
@@ -513,11 +596,14 @@ UploadOperationHandle UploadService::Upload(
    }
 
    auto operation = std::make_shared<AudiocomUploadOperation>(
-      mServiceConfig, fileName, projectName, std::move(completedCallback),
-      std::move(progressCallback));
+      mServiceConfig, fileName, projectName, isPublic,
+      std::move(completedCallback), std::move(progressCallback), trace);
 
-   mOAuthService.ValidateAuth([operation](std::string_view authToken)
-                              { operation->InitiateUpload(authToken); });
+   mOAuthService.ValidateAuth(
+      [operation, this](std::string_view authToken) {
+         operation->InitiateUpload(authToken);
+      },
+      trace, false);
 
    return UploadOperationHandle { operation };
 }
@@ -585,4 +671,4 @@ const auto tempChangedSubscription = TempDirectory::GetTempPathObserver().Subscr
 });
 }
 
-} // namespace cloud::audiocom
+} // namespace audacity::cloud::audiocom
